$(function() {
    var
        // The current group and contact being displayed
        state = { group: null, contact: null, },
        // The list of all groups
        groups = null

    // ----------------------- Application loading ---------------------------

    /**
     * Parse the current browser location (url) for the current user and group.
     * The state variable and application view are updated to reflect the
     * user and group. This is called once when loading the application.
     */
    function restoreState() {
        var currentState = window.location.hash.slice(1),
            all = $("#groups li:first > a");

        state.group = all.data();

        // We are displaying a contact or a contact within a group
        if (/^\/contacts\/\d+$/.test(currentState) ||
            /^\/groups\/[^\/]+\/contacts\/\d+$/.test(currentState)) {
            loadContact(currentState);
            return;
        }
        // We are displaying group
        else if (/^\/groups\/[^\/]+\/contacts$/.test(currentState)) {
            var link = $("#groups a[href='" +
                window.location.hash.replace("'","\\'").replace(" ","%20") +
                "']");
            if (link.length) { link.click(); return }
        }
        // We are displaying the list of all groups (default)
        all.click();
    }

    /**
     * Retrieves the url corresponding to the current view.
     *
     * @return {String} Returns the URL of the current view
     */
    function getState() {
        return      (state.contact && state.contact.href) 
                ||  (state.group && state.group.href)      
                ||  "";
    }

    // ----------------------- Utility functions ---------------------------
    
    /**
     * Show an error message to the user.
     *
     * @param {String} msg The error message to display.
     */
    function showError(msg) {
        var alertdiv = $('<div class="alert alert-error" />')
            .text("Whoops! " + msg + " Please try again later...")
            .prepend($('<a class="close" data-dismiss="alert" data-target="#" href="#' + getState() + '">&times;</a>'));
        alertdiv.alert();
        $("#alerts").append(alertdiv).removeClass('hidden');
    }

    /**
     * A callback function that displays the specified error message if an 
     * error occurred, and calls the specifier callback.
     *
     * @param {String} msg The message to display if an error occurs.
     * @param {Function} callback A callback function.
     */
    function errfn(msg, callback) {
        return function (err) {
            if (msg) showError(msg);
            if (callback) callback(err);
        }
    }

    /**
     * Constructs a callback function that either call the error function if
     * an error occured, or a success function otherwise.
     *
     * @param {Function} efn The function to be called if an error has occured
     * @param {Function} callback The function to be called if no error occured.
     */
    function callbackOrErr(efn, callback) {
        return function(data, err) { if (err) return efn(err); return callback(data); }
    }

    /**
     * Make an AJAX call to the specified URL. Use the specified JQuery
     * parameters. The callback function is called with the retrieved data.
     */
    function ajax(url, jqparms, callback) {
        jqparms.url = url;

        $.ajax(jqparms)
            .done(function (data) { callback(data); })
            .fail(function (jqXHR, msg) { callback(null, msg); });
    }

    /**
     * Constructs the JQuery parameters to be used for sending an update.
     *
     * @param {String} method The HTTP method to be used (PUT, or POST).
     * @param {Object} data A plain JavaScript object, sent to the server as a
     *                      JSON representation.
     */
    function jsonUpdateParms(method, data) {
        return {
            type        : method,
            dataType    : "json",
            contentType : "application/json",
            data        : JSON.stringify(data)
        }
    }

    // ------------------------ Assets management -----------------------------

    /**
     * Retrieves the specified template.
     *
     * @param {String} name The name of the template
     * @param {Function} efn A function called if the template cannot be loaded
     * @param {Function} callback A function called with the loaded template 
     */
    function getTemplate(name, efn, callback) {
        ajax("static/templates/" + name + ".mustache", { dataType: "text", cache:true }, callbackOrErr(efn, callback));
    }

    /**
     * Retrieves the specified json data.
     *
     * @param {String} url The url of the data
     * @param {Function} efn A function called if the data cannot be loaded
     * @param {Function} callback A function called with the loaded data 
     */
    function loadData(url, efn, callback) {
        ajax(url, {dataType: "json", cache: false}, callbackOrErr(efn, callback));
    }

    /**
     * Save the given contact by sending its new JSON representation to the
     * web service.
     */
    function saveContact(newcontact) {
        var isNew = !newcontact.id,
            param = jsonUpdateParms(isNew ? "POST": "PUT", newcontact),
            url= isNew ? "/contacts" : state.contact.href;

        // Send the save request
        ajax(url, param, function(data, err) {
            if (err) { return showError("We could not save this contact information."); }

            $.each(data, function(key, value) { $("#" + key).val(value); });
            $("#remove").removeClass("disabled");

            state.contact = data;
        });
    }

    /**
     * When clicking on the delete button of a contact, remove the contact
     */
    function removeContact(efn, callback) {
        // Are we on the right view, and not creating a new contact
        if (state.contact && state.contact.id) {
            //TODO: remove this contact
            // Call ajax with HTTP method DELETE on the state.contact.href
        }
    }

    // ---------------------- Application controller ---------------------------

    /**
     * Update the list of all groups.
     *
     * @param {Function} callback A callback function called once the list of
     * all groups has been loaded.
     */
    function updateGroups(callback) {
        // Show an error message if the list of groups cannot be loaded
        var efn = errfn("We could not load your groups of contacts.", callback);

        // Try to fetch the list of groups
        loadData("groups/", efn, function(data, err) {
            // Update the application state with the loaded groups
            groups = data;

            // Update the groups to include an "active" function saying whether
            // they are currently selected.
            groups.groups.forEach(function (group) {
                group.active = function() {
                    return (state.group ? (group.href == state.group.href) : false);
                }
            });

            // Try to fetch the groups list template
            getTemplate("groups", efn, function(template, err) {
                // Update the display of the groups
                $("#groups").html(Mustache.to_html(template, data));
                $("#groups a").click(selectGroup);
                if (callback) callback();
            });
        });
    }

    /**
     * Display the specified group.
     *
     * @param {Object} group The group to be displayed. Its {href} attribute
     *                       must contain the url of the group to be loaded.
     * @param {Function} callback A callback function called once the list of
     *                       all groups has been loaded.
     */
    function loadGroup(group, callback) {
        var efn = errfn("We could not load " + (group.group ? "the group '" + group.group + "'." : "your contacts."));

        // Try to load the group data
        loadData(group.href, efn, function(data, err) {
            // Try to the group view template
            getTemplate("contacts", efn, function(template, err) {
                // Unifies the data of the all contacts group with other groups view
                if (!data.contacts) data = {contacts: data};

                // Build the new view
                $("#content").html(Mustache.to_html(template, data));
                $("#content td > a").click(function() {
                    loadContact($(this).data("href"));
                })
                $("#content .dropdown-toggle").dropdown();

                // Update the state
                state = { group: group, contact: null};

                if (callback) callback();
            });
        });
    }

    /**
     * When clicking on a group, update the view accordingly.
     */
    function selectGroup() {
        var target = $(this);
        loadGroup(target.data(), function() {
            $("#groups > li").removeClass("active");
            target.parent().addClass("active");
        });
    }


    /**
     * Add a new group to the contact view.
     *
     * @param {Object} group The group to be added
     * @param {Boolean} contact The contact for which the view is generated
     * @param {Function} callback Called when the display is updated
     */
    function showNewContactGroupEntry(group, contact, callback) {
        var efn = errfn(null, callback);
        getTemplate("contactgroup", efn, function(template_l, err) {
            getTemplate("badge", efn, function(template_b, err) {
                var contactgroup = getContactGroup(group, contact),
                    data = { linkback: getState()
                           , group: group
                           , contacthref: contactgroup ? contactgroup["contact-href"] : null
                           , visible: contactgroup   // Whether the badge is showing
                           },
                    li = $(Mustache.to_html(template_l, data)),
                    badge = $(Mustache.to_html(template_b, data)),
                    a  = $(li).children("a").data("group", group).data("badge",badge).click(toggleGroup);

                $("#contactgroups .contactgroup:last").after(li);
                $("#contactgroupbadges").append(badge);

                if (callback) callback(a);
            });
        });
    }

    /**
     * Display the contact page
     */
    function showContactPage(contact) {
        getTemplate("contact", $.noop, function(template, err) {
            var data = {
                //TODO: add an emailHash field for Gravatar
                contact: contact,
                backlink: state.group && state.group.href,
                };

            state.contact = contact;

            $("#content").html(Mustache.to_html(template, data));

            var grp_back = state.group;
            $("#back").click(function() { loadGroup(grp_back); });
            $("#remove").click(
                function() {
                    var efn = errfn("Whoops... removing that contact did not work!");
                    removeContact(efn, function() { loadGroup(grp_back); })
                });
            $("#creategroupbtn").click(showCreateGroup);
            $("#content input").blur(saveContactFields);

            groups.groups.forEach(function (group) {
                showNewContactGroupEntry(group, contact);
            });
        });
    }

    /**
     * Retrieves a contact from its uri, and display its contact page
     *
     * @param {String} uri The uri at which the representation of the contact
     * can be found.
     */
    function loadContact(uri) {
        loadData(uri, errfn("It seems that we cannot load the specified contact :-/"), showContactPage);
    }

    /**
     * Update the internal representation of the current contact to match the
     * view, and call the web service to save the contact.
     */
    function saveContactFields() {
        var newcontact = {},
            modified = (!state.contact.id);

        // Fetch the new values
        $("#content input").each(function () {
            newcontact[this.id] = this.value;

            modified |= (
                (state.contact[this.id] == null && this.value != "") ||
                (state.contact[this.id] != null && this.value != state.contact[this.id]))
        });

        modified && saveContact(newcontact);
    }

    /**
     * Set the specified group on the contact.
     */
    function setGroup(group, target, callback) {
        ajax(group.href, jsonUpdateParms("POST", state.contact),
            function (data, err) {
                if (err) return;
                state.contact = data;

                target.children("i").addClass("icon-ok").removeClass("icon-empty");
                target.data("badge").removeClass("hidden");

                //Update the contact href
                var contactgroup = getContactGroup(group, state.contact);
                target.data("contacthref", contactgroup["contact-href"]);

                if (callback) callback(data, err);
            });
    }

    /**
     * Remove the specified group when clicking on the corresponding menu entry.
     */
    function unsetGroup(target) {
        ajax(target.data("contacthref"), { type : "DELETE", },
            function (data, err) {
                if (err) return;

                target.children("i").removeClass("icon-ok").addClass("icon-empty");
                target.data("badge").addClass("hidden");

                //Remove the group from the contact
                target.data("contacthref",null);
                state.contact.groups.filter(function(contactgroup) {
                    return (contactgroup.href != target.data("group").href);
                });
            });
    }

    /**
     * Toggle specified group when clicking on the corresponding menu entry.
     */
    function toggleGroup() {
        var target = $(this);

        if (state.contact.id) {
            if (target.data("contacthref")) unsetGroup(target);
            else setGroup(target.data("group"), target);
        }
    }

    function getContactGroup(group, contact) {
        result = null;
        if (contact && contact.groups) {
            contact.groups.forEach(function (contactgroup) {
                if (contactgroup.href == group.href) 
                    result = contactgroup 
            });
        }
        return result;
    }

    /**
     * Find the menu entry that corresponds to the specified group
     */
    function findGroupTarget(group) {
        var result = null;
        $("#contactgroups .contactgroup:not(:first) a").each(function (key, item) {
            if ($(item).data("group").href == group.href) result = $(item);
        });
        return result;
    }

    /**
     * Create a new group, from the group creation dialog box.
     */
    function createGroup() {
        var group = { group: $("#groupname").val() };

        // Send the new group to the web service
        ajax("/groups", jsonUpdateParms("POST", group), function(group, err) {
            // Hide the dialog
            $("#creategroup").modal('hide');
            if (err) return showError("We could not create the group.")

            // If the group did already exist, add it to the current contact
            if (target = findGroupTarget(group)) {
                setGroup(group, target);
            }
            // Otherwise, create a new entry for this group, and update the
            // interface
            else {
                showNewContactGroupEntry(group, null, function(target, err) {
                    if (err) return;
                    setGroup(group, target, function() {
                        updateGroups();
                    });
                });
            }
        });
    }

    /**
     * Display the group creation dialog
     */
    function showCreateGroup() {
        var efn = errfn("The group creation page in unavailable!")
        getTemplate("creategroup", efn, function(template, err) {
            $("#modals").html(Mustache.to_html(template, { linkback: getState() }));
            $("#creategroup").modal();
            $("#groupcreate").click(createGroup);
        });
    }

    $("#newcontact").click(function() { showContactPage({}); });
    updateGroups(restoreState);
});
